
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <netdb.h>
#include <string.h>
#include <netdb.h>
#include <ctype.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <rrd.h>

static u_long packets, rcverr, format, ok, bad;

static void
stats(void)
{
	if (rcverr > 0)
		syslog(LOG_ERR, "recvfrom errors last hour: %lu (total failure would be 3600)", rcverr);	
	rcverr = 0;
	if (format > 0)
		syslog(LOG_ERR, "format errors last hour: %lu", format);	
	format = 0;
	if (bad > 0)
		syslog(LOG_ERR, "bad updates last hour: %lu", bad);	
	bad = 0;

	packets = 0;
	ok = 0;
}

static void
process(u_int32_t ip, char *msg) {
	char **ap, *av[6], *p, *aav[10];
	int ac, i;

	if (msg[0] != '0' || msg[1] != '0') {
		format++;
		return;
	}

	for (ac = 0, ap = av; (*ap = strsep(&msg, " \t")) != NULL;) {
		if (**ap != '\0') {
			ac++;
			if (++ap >= &av[6])
				break;
		}
	}
	if (ac != 4) {
		format++;
		return;
	}
#if 0
	for (p = av[1]; *p; p++)
		if (!isgraph(*p) || *p == '/') {
			format++;
			return;
		}
#endif
	for (p = av[2]; *p; p++)
		if (!isgraph(*p)) {
			format++;
			return;
		}
	for (p = av[3]; *p; p++)
		if (!isgraph(*p)) {
			format++;
			return;
		}
	ac = 0;
	aav[ac++] = "update";
	aav[ac++] = av[1];
	aav[ac++] = "-t";
	aav[ac++] = av[2];
	aav[ac++] = av[3];
	aav[ac] = NULL;
	optind = 0;
	rrd_clear_error();
	i = rrd_update(ac, aav);
fprintf(stderr, "%d\n", i);
	if (!rrd_test_error()) {
		ok++;
		return;
	}
fprintf(stderr, "Rrd: <%s> Template: <%s> Value: <%s> - %s\n", 
	    av[1], av[2], av[3], rrd_get_error());
	syslog(LOG_WARNING, "Rrd: <%s> Template: <%s> Value: <%s> - %s\n", 
	    av[1], av[2], av[3], rrd_get_error());
	bad++;
	fflush(stdout);
}

int
main(int argc, char **argv)
{
	int s, error;
	time_t t, hour;
	char cbuf[BUFSIZ], *port, *host, *cp;
	struct sockaddr_in sins, sind;
	socklen_t sl;
	struct hostent *hp;

	if (argc != 3)
		errx(1, "Usage: %s [host:]port dir", argv[0]);

	memset(&sins, 0, sizeof sins);
	sins.sin_family = AF_INET;
	port = strchr(argv[1], ':');
	if (port != NULL) {
		host = argv[1];
		*port++ = '\0';
	} else {
		host = NULL;
		port = argv[1];
	}
	sins.sin_port = htons(strtoul(port, &cp, 0));
	if (*port == '\0' || *cp != '\0')
		errx(1, "Didn't grok port number");
	if (host == NULL) {
		sins.sin_addr.s_addr = INADDR_ANY;
		host = "ANY";
	} else {
		hp = gethostbyname(host);
		if (hp == NULL) {
			herror(host);
			return (-1);
		}
		memcpy(&(sins.sin_addr.s_addr), hp->h_addr, hp->h_length);
	}

	error = chdir(argv[2]);
	if (error < 0)
		err(1, "chdir");

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0)
		err(1, "socket()");
	
	error = bind(s, (struct sockaddr*)&sins, sizeof(sins));
	if (error < 0)
		err(1, "bind()");

	openlog(argv[0], LOG_PID, LOG_LOCAL1);

#if 1
	daemon(1, 0);
#endif

	syslog(LOG_INFO, "listening to %s:%s", host, port);
	time(&hour);
	hour /= 3600;
	for(;;) {
		sl = sizeof (sind);
		time(&t);
		t /= 3600;
		if (t != hour) {
			stats();
			hour = t;
		}
		error = recvfrom(s, cbuf, sizeof(cbuf), 0, (struct sockaddr *) &sind, &sl);
		if (error < 0) {
			syslog(LOG_ERR, "recvfrom: %m");
			rcverr++;
			sleep(1);
			continue;
		}
		packets++;
		cbuf[error] = '\0';
fprintf(stderr, "<%s>\n", cbuf);
		process(ntohl(sind.sin_addr.s_addr), cbuf);
	}
}
